
BeginPackage["IGES`",{"NURBS`"}]

IGES::usage="Represents an IGES object"
NewIGES::usage="Creates a blank IGES object, ready for entities to be added"
AppendToIGES::usage="Adds a NURBS curve or surface object to the IGES object"
IGESExport::usage="Writes an IGES object to disk"

Begin["Private`"]

IGES/:MakeBoxes[IGES[content__],StandardForm]:=ToBoxes[TableForm[Flatten[List[content]]]]

NewIGES[]:=
	IGES[
		CreateStartSection["Mathematica produced IGES file"],
		CreateGlobalSection[],
		{},
		{},
		{}
	]

AppendLineSuffix[expr_String,code_String,index_?NumberQ]:=
	Block[
		{},
		StringJoin[
			StringPadRight[expr," ",72],
			{code},
			StringPadLeft[index," ",7]
		]
	]

StringPadRight[expr_, char_String:" ", n_:8] := StringJoin[ PadRight[Characters[ToString[expr]],n,char] ] 

StringPadLeft[expr_, char_String:" ", n_:8] := StringJoin[ PadLeft[Characters[ToString[expr]],n,char] ]

Hollerith[string_String] := StringJoin[ ToString[StringLength[string]],"H",string ]

CreateStartSection[msg_String]:=
	Block[
		{},
		List[List[AppendLineSuffix[msg,"S",1]]]
	]

Options[CreateGlobalSection]=
	{
		ProductID->"None",
		FileID->"None",
		DoublePrecisionSignificance->Floor[$MachinePrecision],
		DoublePrecisionMagnitude->Floor[Log[10, $MaxMachineNumber]],
		NativeBits->$SystemWordLength,
		Resolution->0.001,
		UserName->$UserName
	};


CreateGlobalSection[options:OptionsPattern[]]:=
	Block[
		{list},
		list=
			ParameterAssemble[{
				",", 								(* parameter delimiter *)
				";", 							(*	 record delimieter *)
				OptionValue[ProductID], 	(* product id string *)
				OptionValue[FileID], 			(* file name *)
				"NURBS Package Utility", 	(* native system id *)
				"1.0", 							(* preprocessor version *)
				OptionValue[NativeBits], 								(* native system integer bits *)
				38, 								(* single precision magnitude *)
				6, 								(* single precision significance *)
				OptionValue[DoublePrecisionMagnitude], 		(* double precision magnitude *)
				OptionValue[DoublePrecisionSignificance], 		(* double precision significance *)
				"None", 							(* product id for receiver *)
				1.0, 								(* model space scale *)
				2, 								(* units enumeration 2=mm *)
				"mm", 							(* units name *)
				1, 								(* max number of line weight graduations *)
				0.254, 							(* max line width *)
				DateString[{"Year","Month","Day",".","Hour24","Minute","Second"}], (* creation date *)
				OptionValue[Resolution], 							(* user intended resolution *)
				0.0, 								(* model bounding box size, or zero *)
				OptionValue[UserName], 							(* name of author *)
				11, 								(* IGES version enumeration to which ths file complies 11=5.3 *)
				0, 								(* drafting standard *)
				DateString[{"Year","Month","Day",".","Hour24","Minute","Second"}] (* native model modification date *)
			}];
		Return[List[MapIndexed[AppendLineSuffix[#1,"G",First[#2]]&,list]]];
]


Options[DirectoryEntry]=
	{
		LineFont->Automatic,
		LineWeight->Automatic,
		Status->Automatic,
		Color->Automatic
	}


DirectoryEntry[entity_, paramptr_, structureptr_, level_, view_, transformptr_, labelptr_, paramlinecount_,
	formnumber_, label_, subscript_, dirindex_, options:OptionsPattern[]]:=
	Block[
		{dirlines,linefont,status,lineweight,color},
		linefont=OptionValue[LineFont]/.Automatic->0;
		status=OptionValue[Status]/.Automatic->0;
		lineweight=OptionValue[LineWeight]/.Automatic->0;
		color=OptionValue[Color]/.Automatic->0;

		dirlines=List[
			StringJoin[
				StringPadLeft[entity],
				StringPadLeft[paramptr],
				StringPadLeft[structureptr],
				StringPadLeft[linefont],
				StringPadLeft[level],
				StringPadLeft[view],
				StringPadLeft[transformptr],
				StringPadLeft[labelptr],
				StringPadLeft[status,"0"]
			],
			StringJoin[
				StringPadLeft[entity],
				StringPadLeft[lineweight],
				StringPadLeft[color],
				StringPadLeft[paramlinecount],
				StringPadLeft[formnumber],
				StringPadLeft[0],
				StringPadLeft[0],
				StringPadRight[label],
				StringPadLeft[subscript]
			]
		];

		MapIndexed[
			AppendLineSuffix[#1,"D",First[#2]+dirindex-1]&,
			dirlines
		]
	]


ParameterAssemble[params_List,n_:72,options:OptionsPattern[]]:=
	Block[
		{linelist,line,element},
		linelist={};
		element=ParamString[params[[1]]];
		line=element;

		For[i=2,i<=Length[params],i++,
			element=ParamString[params[[i]]];
			If[
				n-StringLength[line]-1>StringLength[element],
				line=StringJoin[line,",",element],
				line=StringJoin[line,","];
				AppendTo[linelist,line];
				line=element
			];
		];
		line=StringJoin[line,";"];
		AppendTo[linelist,line];
		Return[linelist];
	]


ParamString[expr_String] := Hollerith[expr]

ParamString[expr_?IntegerQ] := ToString[expr]

ParamString[expr_?NumberQ] := ToString[ NumberForm[expr,6] ]


ParameterRecord[entity_?NumberQ,params_List,dirindex_?NumberQ,paramindex_?NumberQ]:=
	Block[
		{linelist},
		linelist=ParameterAssemble[Join[{entity},params],64];

		linelist=
		Map[
			StringJoin[StringPadRight[#1," ",64],StringPadLeft[dirindex,"0",8]]&,
			linelist
		];
		MapIndexed[
			AppendLineSuffix[#1,"P",First[#2]+paramindex-1]&,
			linelist
		]
	]


AddTerminateSection[iges_IGES]:=
	Block[
		{igesobj},
		igesobj=iges;

		igesobj[[5]]=
			AppendLineSuffix[
				StringPadRight[
					StringJoin[
						"S",StringPadLeft[Length[igesobj[[1]]],"0",7],
						"G",StringPadLeft[Length[igesobj[[2]]],"0",7],
						"D",StringPadLeft[Length[igesobj[[3]]],"0",7],
						"P",StringPadLeft[Length[igesobj[[4]]],"0",7]
					]," ",72],
			"T",1];

		Return[igesobj];
	]


Attributes[AppendToIGES]={HoldFirst}

AppendToIGES[iges_,curve_NURBSpline]:=
	Block[
		{dirseq,dirlines,paramseq,paramlines,entitynum},
		entitynum=126;
		dirseq=Length[iges[[3]]]+1;
		paramseq=Length[iges[[4]]]+1;

		paramlines=
			ParameterRecord[
				entitynum,
				Join[
					{Length[HPolygon[curve]]-1,CurveDegree[curve],0,0,1,0}, (* number of points-1, dimension of points, flags *)
					N@Knots[curve],
					N@Weights[curve],
					Flatten[Euclidean[HPolygon[curve]]], (* flattened list of points *)
					{0,1}, (* parameter range *)
					{0,0,1} (* unit normal, for planar curve case *)
				],
				dirseq,paramseq
			];

		dirlines=DirectoryEntry[entitynum,paramseq,0,1,0,0,0,Length[paramlines],0,"Curve",0,dirseq];

		iges[[4]]=Join[iges[[4]],paramlines];
		iges[[3]]=Join[iges[[3]],dirlines];
		iges=AddTerminateSection[iges];
		Return[];
	]

AppendToIGES[iges_,surface_NURBSSurface] :=
	Block[
		{dirseq,dirlines,paramseq,paramlines,entitynum},
		entitynum=128;
		dirseq=Length[iges[[3]]]+1;
		paramseq=Length[iges[[4]]]+1;

		Catch[
			paramlines=
				ParameterRecord[
				entitynum,
				Flatten[
					{
						Dimensions[HPolygon[surface][[All,All,1]]]-1, (* upts-1, vpts-1 *)
						SurfaceDegree[surface],
						{0,0,1,0,0},
						N@UKnots[surface],
						N@VKnots[surface],
						N@Weights[surface],
						Euclidean[HPolygon[surface]], (* flattened list of points *)
						{0,1}, (* u parameter range *)
						{0,1} (* v parameter range *)
					}
				],
				dirseq,paramseq];
	
			dirlines=DirectoryEntry[entitynum,paramseq,0,1,0,0,0,Length[paramlines],0,"bsSurf",0,dirseq];
	
			iges[[4]]=Join[iges[[4]],paramlines];
			iges[[3]]=Join[iges[[3]],dirlines];
			iges=AddTerminateSection[iges];
			Return[];
		]
	]


IGESExport[filename_String, object_]:=
	Block[
		{iges,lines},
		iges=NewIGES[];
		If[Head[object]===List,MapIndexed[AppendToIGES[iges,#1]&,object],AppendToIGES[iges,object]];
		lines=Flatten[iges/.IGES->List];
		Export[filename,lines,"List"]		
	]
	



End[]


EndPackage[]




