(* NURBS Package source *)

BeginPackage["NURBS`"]
(* Exported symbols added here with SymbolName::usage *) 

(* Basis Functions *)
Basis::usage="Basis[i,p,U] gives the ith Cox-deBoor basis function of degree p, given the knot vector U."
SplineSpace::usage="SplineSpace[p,U] gives the piecewise defined vector of symbolic Cox-deBoor basis functions."

(* Homogeneous coordinates *)
Homogeneous::usage="Homogeneous[point] gives the homogeneous coordinate representation of point.  Homogeneous[array] gives the homogeneous representation of a list of points in array."
Euclidean::usage=""

(* Knot vectors *)
KnotVector::usage="KnotVector[n] gives a knot vector of length n.  An open, uniform knot vector is given by default.  KnotVector[points] gives a knot vector for the list of points points."
KnotKernel::usage="An option to KnotVector that specifies the type of knots to return."
UniformClamped::usage="A value for KnotKernel that specifies an open, uniform knot vector."
Uniform::usage="A value of KnotKernel that specifies a periodic knot vector."
Averaging::usage="A value of KnotKernel that specifies a parameter averaged knot kernel.  Requires passing the parameters as a list via the ParameterVector option."

(* Parameterization *)
ParameterVector::usage="An option to KnotVector that gives the parameterization, needed by the Averaging knot kernel"
Parameterization::usage="An option to various fitting functions that specifies the parameterization to be used."
ChordLength::usage="A value of Parameterization that specifies chord length parameterization."
Centripetal::usage="A value of Parameterization that specifies centripetal parameterization."

(* options to fitting functions *)
ControlPoints::usage="An option to approximation functions that specifies the number of control points to be returned."
CurveDegree::usage="An option to curve approximation and interpolation functions that specifies the degree."
PConstraints::usage="Option to NURBSplineApproximate[] which determines which points in the point list are to be constrained."
DConstraints::usage="Option to NURBSplineApproximate[] which determines which derivatives in the given list are to be constrained."
SurfaceDegree::usage="An option to surface approximation and interpolation functions that specifies the degree.  It must be in the form {p,q}, where p and q are the degrees in the u- and v-directions, respectively."
ParameterizationIndex::usage="An option to surface approximation and interpolation functions that specifies which u-direction and v-direction rows to use for calculation of parameters."

(* Property accessors *)
HPolygon::usage="HPolygon[curve] or HPolygon[surface] returns the control points for the curve or surface."
Weights::usage="Weights[curve] or Weights[surface] returns the control point weights for the curve or surface."
Knots::usage="Knots[curve] returns the knot vector for the curve."
UKnots::usage=""
VKnots::usage=""

(* Test *)
NURBSQ::usage = "NURBSQ  "

(* Spline object and evaluation *)
NURBSpline::usage="Represents a spline object."
SplineFunction::usage="SplineFunction[curve] returns a function that evaluates to the location on the curve for the given parameter."

(* Spline object construction *)
NURBSplineFromPolygon::usage="NURBSplineFromPolygon[points] returns a spline object with points as the control points."
NURBSplineInterpolate::usage="NURBSplineInterpolate[points] returns a spline object that interpolates points."
NURBSplineApproximate::usage="NURBSplineApproximate[points,derivs] gives a spline object that fits the given points approximately.  Points are given in the format {x,y[,z]} and derivatives are given in the format {d1,Null,Null,...,dn}."
NURBSplineApproximate::ovrcons="The system is overconstrained."

(* Surface object and evaluation *)
NURBSSurface::usage="Represents a B-spline surface object."
SurfaceFunction::usage="SurfaceFunction[surface] returns a function that evaluates to the location on the surface for the given parameters."

(* Surface object construction *)
NURBSSurfaceFromPolygon::usage="NURBSSurfaceFromPolygon[points] returns a surface object with 2-D array of points points as the control points."
NURBSSurfaceInterpolate::usage="NURBSSurfaceInterpolate[points] returns a surface object that interpolates the 2-D array of points points.  NURBSSurfaceInterpolate[f,{m,n}] returns a surface object that interpolates points sampled from function f in an m by n array."
NURBSSurfaceFitError::usage="NURBSSurfaceFitError[f,surface] returns and array of differences between f and surface."










Begin["`Private`"]
(* Implementation of the package *)

Homogeneous[array_?ArrayQ] := 
	Map[
		Append[#1,1]&,
		array,
		{ArrayDepth[array]-1}
	]

Euclidean[array_?ArrayQ] :=
	Map[
		Delete[#1/#1[[-1]],-1]&,
		array,
		{ArrayDepth[array]-1}
	]
	
NURBSQ[object_]:=MemberQ[{NURBSpline,NURBSSurface},Head[object]]

ZRatio[a_, 0] := If[a!=0,Message[ZRatio::infy,StringJoin[ToString[a],"/0"]],0]

ZRatio[a_, b_] := a/b

FindSpan[U_]:=
	Module[
		{K,z},
		K=Union[U];
		z=(Length[U]-Length[K])/2-1;
		Function[{u},Interpolation[Transpose[{K,Range[Length[K]]}],InterpolationOrder->0][u]+z]
	]

Options[Basis]=
	{
		KnotSpan->Automatic
	};

Basis[i_, 0,knots_List,options:OptionsPattern[]] :=
	Module[
		{span},
		span=Function[u, OptionValue[KnotSpan]/.{Automatic:>FindSpan[knots][u]}];

		Function[u, If[i == span[u], 1, 0] ]
	]

Basis[i_,p_,knots_List,options:OptionsPattern[]] :=
	Module[
		{span},
		span=Function[u,OptionValue[KnotSpan]/.{Automatic:>FindSpan[knots][u]} ];

		Function[u, 
			ZRatio[Basis[i,p-1,knots,KnotSpan->span[u]][u](u-knots[[i]]),knots[[i+p]]-knots[[i]]]
			+ZRatio[Basis[i+1,p-1,knots,KnotSpan->span[u]][u]( knots[[i+p+1]]-u),knots[[i+p+1]]-knots[[i+1]]] 
		]
	]

nz[a_,b_]:=If[b==0,0,a/b]

BasisD[i_,p_,knots_List,options:OptionsPattern[]]:=
	Function[u,
		Basis[i,p-1,knots,options][u] * nz[p,(knots[[i+p]]-knots[[i]])] 
		- Basis[i+1,p-1,knots,options][u] * nz[p,(knots[[i+p+1]]-knots[[i+1]])]
	]

SplineSpace[p_,U_]:=
	Block[
	{m,zvec},
	m=Length[U]-p-2;
	zvec=ConstantArray[0,{m+1}];

	Function[{u},
		Evaluate[
			Piecewise[
				Join[
					Table[
						{

						ReplacePart[zvec,
							Table[
								{i}->Collect[Basis[i,p,U,KnotSpan->j][u],u], (* use Collect[] to speed evaluation *)
								{i,j-p,j} (* nonzero basis functions for this knot span *)
							]
						],

						U[[j]]<=u<U[[j+1]] (* piecewise condition *)

						},
						{j,p+1,m+1} (* interior knot spans *)
					],
					{{
						ReplacePart[zvec,
							Table[
								{i}->Collect[Basis[i,p,U,KnotSpan->m+1][1],u], (* use Collect[] to speed evaluation *)
								{i,m-p+1,m+1} (* nonzero basis functions for this knot span *)
							]
						], (* piecewise endpoint case value *)
						u==U[[m+2]] (* piecewise endpoint condition *)
					}}
					],zvec (* the default value for Piecewise[] must be vector valued for the function to compile *)
			]
		]
	]
	]

Options[KnotVector] = 
	{
		KnotKernel -> "UniformClamped",
		CurveDegree -> 3,
		ParameterVector->None
	};

KnotVector[length_?NumberQ, options:OptionsPattern[]] :=
	Block[
		{p, n,KnotKernel},
		KnotKernel=ToExpression[OptionValue[KnotKernel]];
		p = OptionValue[CurveDegree];
		n = length - p - 2;
		Table[
			KnotKernel[i,n,p,options],
			{i, 1, length}
		]
	]

KnotVector[vertices_List, options:OptionsPattern[]] :=
	Block[
		{p},
		p = OptionValue[CurveDegree];
		KnotVector[Length[vertices] + p + 1, options]
	]

UniformClamped[i_, n_, p_, options:OptionsPattern[]] :=
	Piecewise[{
		{0, i<=(p+1)},
		{(i-p-1)/(n-p+1), (p+1)< i <=(n+1)},
		{1, i>(n+1)}
	}]

Uniform[i_, n_, p_, options:OptionsPattern[]] := (i-p-1)/(n-p+1)

Options[Averaging]=Options[KnotVector];

Averaging[i_, n_, p_, options:OptionsPattern[]]:=
	Block[
		{K},
		K=OptionValue[ParameterVector]; (* TODO: argument checking *)
		Piecewise[{
			{0, i<=(p+1)},
			{Sum[K[[j]], {j,i-p-1,i-2}]/p, (p+1)<i<=(n+1)},
			{1, i>(n+1)}
		}]
	]

ChordLength[points_List,degree_?NumberQ]:=
	Module[
		{P,n,dist},
		dist[pt1_,pt2_]:=Norm[pt2-pt1];
		n=Length[points]-1;
		P=Table[
			dist[points[[i]],points[[i+1]]],
			{i,1,n}
		];
		PrependTo[P,0];
		Accumulate[P]/Total[P]
	]

Centripetal[points_List,degree_?NumberQ]:=
	Module[
		{P,n,dist},
		dist[pt1_,pt2_]:=Sqrt[Norm[pt2-pt1]];
		n=Length[points]-1;
		P=Table[
			dist[points[[i]],points[[i+1]]],
			{i,1,n}
		];
		PrependTo[P,0];
		Accumulate[P]/Total[P]
	]

NURBSpline[hpolygon_List, knots_List, degree_?NumberQ];

NURBSpline/:MakeBoxes[NURBSpline[hpolygon_List, knots_List,degree_?NumberQ],StandardForm]:= ToBoxes[NURBSpline[Shallow[hpolygon,2],Shallow[knots,2],degree]]

NURBSpline/:CurveDegree[NURBSpline[hpolygon_List, knots_List, degree_?NumberQ]] := degree

NURBSpline/:Knots[NURBSpline[hpolygon_List, knots_List, degree_?NumberQ]] := knots

NURBSpline/:HPolygon[NURBSpline[hpolygon_List, knots_List, degree_?NumberQ]] := hpolygon

NURBSpline/:Weights[NURBSpline[hpolygon_List, knots_List, degree_?NumberQ]] := hpolygon[[All,-1]]

Options[NURBSplineFromPolygon] = 
	{
		CurveDegree->Automatic, 
		KnotKernel->"OpenUniform",
		Closed->False
	};

NURBSplineFromPolygon[vertices_List, options:OptionsPattern[]] :=
	Block[
		{p,m,r,U,polygon},
		p=OptionValue[CurveDegree]/.Automatic->3;
		If[OptionValue[Closed],polygon=Join[vertices,vertices[[;;p]]],polygon=vertices];
		U=KnotVector[polygon,FilterRules[{options},Options[KnotVector]]];
		Return[NURBSpline[Homogeneous[polygon],U,p]];
	]

SplineFunction[spline_NURBSpline,function_:Automatic] :=
	Block[
		{Bu,polygon,p,U,m},
		polygon=Euclidean[HPolygon[spline]];
		p=CurveDegree[spline];
		U=Knots[spline];
		m=Length[polygon]-1;

		Bu=function/.Automatic->SplineSpace[p,U];
		Function[u,
			Evaluate[
			Bu[u].polygon
			]
		]
	]

NURBSpline/:Graphics[spline_NURBSpline,options:OptionsPattern[]]:=
	Block[
		{f},
		f=SplineFunction[spline];
		Show[
			ParametricPlot[
				f[u],
				{u,0,1},
				PerformanceGoal->"Speed",
				AspectRatio->Automatic,
				Axes->None
			],
			ListPlot[
				Euclidean[HPolygon[spline]],
				Joined->True,
				PlotStyle->LightGray
			],
			options
		]
	]

NURBSpline/:Graphics3D[spline_NURBSpline,options:OptionsPattern[]]:=
	Block[
		{f},
		f=SplineFunction[spline];
		Show[
			ListPointPlot3D[
				Euclidean[HPolygon[spline]],
				PlotStyle->LightGray,
				Axes->None
			],
			ParametricPlot3D[
				f[u],
				{u,0,1},
				PerformanceGoal->"Speed"
			],
			options
		]
	]

Options[NURBSplineInterpolate]=
	{
		CurveDegree->3,
		Parameterization->"ChordLength",
		KnotKernel->"OpenUniform"
	};

NURBSplineInterpolate[points_List,options:OptionsPattern[]] := 
	Module[
		{m,n,p,M,U,K,polygon},
		n=Length[points]-1;
		p=OptionValue[CurveDegree]/.Automatic->3;
		m=n+p+1;
		K=ToExpression[OptionValue[Parameterization]][points,p]; (* generate parameter list *)
		U=KnotVector[m+1,Join[{ParameterVector->K},FilterRules[{options}, Options[KnotVector]]]];

		M=Table[
			Piecewise[{
				{1,j==1&&i==1},
				{Basis[j,p,U][K[[i]]],1<i<n+1},
				{1,j==n+1&&i==n+1}
			}],
			{i,1,n+1},{j,1,n+1}];
		polygon=N[LinearSolve[M,points]];
		NURBSpline[Homogeneous[polygon],U,p]
	]

prvSplineInterpCoeffArrayD[n_,p_,U_,K_]:=
	Block[
		{M},
		M=Table[ (* populate interior point equations *)
			Piecewise[{
				{Basis[j,p,U][K[[i-1]]],3<=i<=n+1} (* for the first interior equation, i=3, K[[2]] is the first interesting parameter *)
			}],
			{i,1,n+3},{j,1,n+3}
		];
		M[[1,1]]=1; (* multipliers for endpoint constraints *)
		M[[-1,-1]]=1;
		M[[2,1]]=-1;M[[2,2]]=1; (* multipliers for derivative constraints *)
		M[[-2,-2]]=-1;M[[-2,-1]]=1;
		Return[M];
	]

prvSplineInterpDataVectorD[points_,derivs_,n_,p_,U_]:=
	Block[
		{Q},
		Q=points;
		Q=Insert[Q,U[[p+2]]*derivs[[1]]/p,2]; (* derivative vector data inserted *)
		Q=Insert[Q,(1-U[[n+3]])*derivs[[-1]]/p,-2];
		Return[Q];
	]

NURBSplineInterpolate[points_List,derivs_List,options:OptionsPattern[]] := 
	Module[
		{m,n,p,M,U,K,Q,polygon},
		n=Length[points]-1;
		p=OptionValue[CurveDegree]/.Automatic->3;
		m=n+p+3; (* pick up two additional control points due to end derivative spec *)
		K=ToExpression[OptionValue[Parameterization]][points,p]; (* generate parameter list *)
		U=KnotVector[m+1,Join[{ParameterVector->K}, FilterRules[{options},Options[KnotVector]]]];

		M=prvSplineInterpCoeffArrayD[n,p,U,K];
		Q=prvSplineInterpDataVectorD[points,derivs,n,p,U];
		polygon=N@LinearSolve[M,Q];
		NURBSpline[Homogeneous[polygon],U,p]
	]

NURBSplineInterpolate[points_List/;Depth[points]==4, derivs_List/;Depth[derivs]==4, options:OptionsPattern[]] := 
	Module[
		{m,n,p,M,U,K,Q,polygon,getpoly},
		n=Length[points[[1]]]-1;
		p=OptionValue[CurveDegree]/.Automatic->3;
		m=n+p+3; (* pick up two additional control points due to end derivative spec *)
		K=ToExpression[OptionValue[Parameterization]][points[[1]],p]; (* generate parameter list *)
		U=KnotVector[m+1,FilterRules[{options},Options[KnotVector]]];

		M=prvSplineInterpCoeffArrayD[n,p,U,K];
		getpoly[pvec_,dvec_]:=
			Block[
			{},
			Q=prvSplineInterpDataVectorD[pvec,dvec,n,p,U];
			N@LinearSolve[M,Q]
		];
		polygon=MapThread[getpoly,{points,derivs},1];

		MapThread[NURBSpline[Homogeneous[#1],U,p]&,polygon,1]
	]

Options[NURBSplineApproximate]=
	{
		CurveDegree->Automatic,
		ControlPoints->Automatic,
		PConstraints->{1,-1},
		DConstraints->{1,-1},
		Parameterization->"ChordLength",
		KnotKernel->"UniformClamped"
	};
	
NURBSplineApproximate::cpns="Warning: Number of control points not specified.  Defaulting to `1`"
	
NURBSplineApproximate[points_,derivs_,options:OptionsPattern[]]:=
	Block[
		{p,n,m,S,T,Bu,Bc,U,K,spline},
		(* p=curve degree *)(* n=number of control points *) (* m=number of data points *)
		(* S=unconstrained items *)(* T=constrained items *)
		(* P=unknown control pts *)(* Bu=basis or basis' matrix *)
		(* Bc=basis or basis' matrix *)
		(* U=knot vector *)(* K=parameter list *)
		(* spline=spline object *)
		(* TODO: argument checking: pts is a list, drs is a list, length of lists must be th same *)
		(* TODO: check for underconstrained or poorly conditioned systems *)
		(* TODO: issue a warning when ControlPoints is not specified *)
		p=OptionValue[CurveDegree]/.Automatic->3;
		If[OptionValue[ControlPoints]===Automatic,Message[NURBSplineApproximate::cpns,p+1]];
		n=OptionValue[ControlPoints]/.Automatic->p+1;
		m=Length[points];
		U=KnotVector[n+p+1,FilterRules[{options},Options[KnotVector]]];
		K=ToExpression[OptionValue[Parameterization]][points,p]; (* generate parameter list *)
		
		Catch[
			If[Length[OptionValue[PConstraints]]+Length[OptionValue[DConstraints]]>n,Throw[Message[NURBSplineApproximate::ovrcons]]];
	
			Block[
				{jrange1,jrange2},
	
				jrange1=Delete[Range[m],Map[List,OptionValue[PConstraints],1]];
				jrange2=Delete[Flatten@Position[derivs,_List,{1}],Map[List,OptionValue[DConstraints]]];
				Bu=Join[ (* basis functions for unconstrained points and derivatives *) (* join matrices row-wise *)
					Table[Basis[i,p,U][K[[j]]],{j,jrange1},{i,1,n}],(* basis functions *)
					Table[BasisD[i,p,U][K[[j]]],{j,jrange2},{i,1,n}] (* basis function derivatives *)
				];
				S=Join[
					Part[points,jrange1], (* unconstrained points *)
					Part[derivs,jrange2] (* unconstrained derivatives *)
				];
	
				jrange1=OptionValue[PConstraints];
				jrange2=OptionValue[DConstraints];
				Bc=Join[ (* position constrained points and derivatives *)
					Table[Basis[i,p,U][K[[j]]],{j,jrange1},{i,1,n}],(* basis functions *)
					Table[BasisD[i,p,U][K[[j]]],{j,jrange2},{i,1,n}] (* basis function derivatives *)
				];
				T=Join[
					Part[points,jrange1], (* constrained points *)
					Part[derivs,jrange2] (* constrained derivatives *)
				];
			];
	
			M=ArrayFlatten[{{Transpose[Bu].Bu,Transpose[Bc]},{Bc,0}}];
			B=Join[Transpose[Bu].S,T];
			spline=NURBSplineFromPolygon[LinearSolve[M,B][[1;;n]]];

			Return[spline];
		
		];
]

NURBSSurface[polygon_List,{uknots_List,vknots_List},{udegree_,vdegree_}];

NURBSSurface/:MakeBoxes[NURBSSurface[polygon_List,{uknots_List,vknots_List},{udegree_,vdegree_}],StandardForm]:=
ToBoxes[NURBSSurface[StringJoin["\[LeftSkeleton]",ToString[Dimensions[polygon]],"\[RightSkeleton]"],StringJoin["\[LeftSkeleton]",ToString[Dimensions[{uknots,vknots}]],"\[RightSkeleton]"],{udegree,vdegree}]]

NURBSSurface/:HPolygon[NURBSSurface[polygon_List,{uknots_List,vknots_List},{udegree_,vdegree_}]]:=polygon

NURBSSurface/:CurveDegree[NURBSSurface[polygon_List,{uknots_List,vknots_List},{udegree_,vdegree_}]]:={udegree,vdegree}

NURBSSurface/:UKnots[NURBSSurface[polygon_List,{uknots_List,vknots_List},{udegree_,vdegree_}]]:=uknots

NURBSSurface/:VKnots[NURBSSurface[polygon_List,{uknots_List,vknots_List},{udegree_,vdegree_}]]:=vknots

NURBSSurface/:SurfaceDegree[NURBSSurface[polygon_List,{uknots_List,vknots_List},{udegree_,vdegree_}]]:={udegree,vdegree}

NURBSSurface/:Weights[NURBSSurface[polygon_List,{uknots_List,vknots_List},{udegree_,vdegree_}]]:=polygon[[All,All,-1]]

Options[NURBSSurfaceFromPolygon]=
	{
		SurfaceDegree->{3,3}
	};

NURBSSurfaceFromPolygon[vertices_List,options:OptionsPattern[]]:=
	NURBSSurface[
		Homogeneous[vertices],
		{
			KnotVector[vertices[[All,1]],FilterRules[{options},Options[KnotVector]]],
			KnotVector[vertices[[1,All]],FilterRules[{options},Options[KnotVector]]]
		},
		OptionValue[SurfaceDegree]
	]

SurfaceFunction[surface_NURBSSurface,ufunction_:Automatic,vfunction_:Automatic] :=
	Block[
		{polygon,p,q,U,V,m,n,Bu,Bv},
		polygon = Euclidean[HPolygon[surface]];
		{p,q} = SurfaceDegree[surface];
		U=UKnots[surface];
		V=VKnots[surface];
		{m,n}=Dimensions[polygon][[1;;2]]-1;
		Bu=ufunction/.Automatic->SplineSpace[p,U];
		Bv=vfunction/.Automatic->SplineSpace[q,V];

		Compile[{u,v},
			Evaluate[
				Table[
					Bu[u].polygon[[All,All,i]].Bv[v],
					{i,1,3}
				]
			]
		]
	]

NURBSSurface/:Graphics3D[surface_NURBSSurface,options:OptionsPattern[]]:=
	Block[
		{f},
		f=SurfaceFunction[surface];
		Show[
			ParametricPlot3D[
				f[u,v],
				{u,0,1},
				{v,0,1},
				Mesh->{Union[UKnots[surface]],Union[VKnots[surface]]},
				ColorFunction->"GrayTones"
			],
			ListPointPlot3D[
				Euclidean[HPolygon[surface]],
				PlotStyle->LightGray
			],
			PlotRange->All,
			Axes->None,
			Boxed->False
		]
	]

Options[NURBSSurfaceInterpolate]=
	{
		SurfaceDegree->Automatic,
		Parameterization->"ChordLength",
		KnotKernel->"OpenUniform",
		ParameterizationIndex->Automatic
	};

(* BUG: calculated surface does not pass through given points *)
NURBSSurfaceInterpolate[points_List,options:OptionsPattern[]]:=
	Block[
		{A,B,p,q,m,n,r,s,w,y,K,L,U,V,polygon},
		{m,n}=Dimensions[points][[1;;2]]-1;
		{p,q}=OptionValue[SurfaceDegree]/.Automatic->{3,3};
		{r,s}=OptionValue[ControlPoints]/.Automatic->{m+p+1,n+q+1};
		{w,y}=OptionValue[ParameterizationIndex]/.Automatic->{1,1};
		K=ToExpression[OptionValue[Parameterization]][points[[All,w]],p]; (* generate parameter list *)
		L=ToExpression[OptionValue[Parameterization]][points[[y,All]],q];
		U=KnotVector[r+1,FilterRules[{options},Options[KnotVector]]]; (* TODO: refine knot positions using parameters *)
		V=KnotVector[s+1,FilterRules[{options},Options[KnotVector]]];
		A=Table[
			Basis[i,p,U][K[[j]]],
			{j,1,m+1},{i,1,m+1} (* j rows and i columns *)
		];
		B=Table[
			Basis[i,q,V][L[[j]]],
			{i,1,n+1},{j,1,n+1} (* i rows and j columns *)
		];
		polygon=Transpose[A.Inner[Times,points,B,Plus,2],{1,3,2}]; (* A^-1 Q B^-1 , allowing for the elements of points to be vector valued *)
		NURBSSurface[Homogeneous[polygon],{U,V},{p,q}]
	]

NURBSSurfaceInterpolate[function_,{upts_,vpts_},options:OptionsPattern[]]:=
	Block[
		{points,params,derivs,p,q,m,n,r,s,w,y,K,L,U,V,M,Q,polygon},
		points=Table[
			N@function[(i-1)/upts,(j-1)/vpts],
			{i,1,upts+1},{j,1,vpts+1}
		];
		params=Table[
			{(i-1)/upts,(j-1)/vpts},
			{i,1,upts+1},{j,1,vpts+1}
		];
		{m,n}=Dimensions[points][[1;;2]]-1;
		{p,q}=OptionValue[SurfaceDegree]/.Automatic->{3,3};
		{r,s}={m+p+3,n+q+3}; (* just like spline fitting, we pick up two extra control points due to derivative constraints *)
		{w,y}=OptionValue[ParameterizationIndex]/.Automatic->{1,1};
		K=ToExpression[OptionValue[Parameterization]][points[[All,w]],p]; (* generate parameter list *)
		L=ToExpression[OptionValue[Parameterization]][points[[y,All]],q];
		U=KnotVector[r+1,Join[{ParameterVector->K},FilterRules[{options},Options[KnotVector]]]]; (* TODO: refine knot positions using parameters *)
		V=KnotVector[s+1,Join[{ParameterVector->L},FilterRules[{options},Options[KnotVector]]]];

		
		(* first direction fitting pass walks in u-direction with curves in v-direction *)
		derivs=ConstantArray[0,{upts+1,2}];
		Do[(* endpoint derivatives *)
			derivs[[i,1]]=N@D[function[u,v],v]/.{u->(i-1)/(upts),v->0};
			derivs[[i,2]]=N@D[function[u,v],v]/.{u->(i-1)/(upts),v->1},
			{i,1,upts+1}
		];

		M=prvSplineInterpCoeffArrayD[n,q,V,L];
		polygon=Table[
			Q=prvSplineInterpDataVectorD[points[[i,All]],derivs[[i]],n,q,V];
			LinearSolve[M,Q],
			{i,1,upts+1}
		];

		(*  second direction pass *)
		derivs=ConstantArray[0,{vpts+1,2}];
		Do[(* endpoint derivatives *)
			derivs[[i,1]]=N@D[function[u,v],u]/.{v->(i-1)/(vpts),u->0}; 
			derivs[[i,2]]=N@D[function[u,v],u]/.{v->(i-1)/(vpts),u->1},
			{i,1,vpts+1}
		];

		(* need to add two new stations for the additional rows of control points *)
		(* stuff two new derivative pairs - interpolate chord-wise for new stations *)
		Block[
			{v0,v1,dist},
			dist[x_,y_]:=Sqrt[(x-y).(x-y)]; (* distance utility *)
			v0=params[[1,1,2]]+(dist[polygon[[1,1]],polygon[[1,2]]]/dist[polygon[[1,1]],polygon[[1,3]]])*(params[[1,2,2]]-params[[1,1,2]]);
			v1=params[[-1,1,2]]+(dist[polygon[[-1,1]],polygon[[-1,2]]]/dist[polygon[[-1,1]],polygon[[-1,3]]])*(params[[-1,2,2]]-params[[-1,1,2]]);
			derivs=Insert[ (* near v=0 edge *)
				derivs,
				{
					N@D[function[u,v],u]/.{v->v0,u->0},
					N@D[function[u,v],u]/.{v->v1,u->1}
				},
				2
			];
			v0=params[[1,-1,2]]+(dist[polygon[[1,-1]],polygon[[1,-2]]]/dist[polygon[[1,-1]],polygon[[1,-3]]])*(params[[1,-2,2]]-params[[1,-1,2]]);
			v1=params[[-1,-1,2]]+(dist[polygon[[-1,-1]],polygon[[-1,-2]]]/dist[polygon[[-1,-1]],polygon[[-1,-3]]])*(params[[-1,-2,2]]-params[[-1,-1,2]]);
			derivs=Insert[ (* near v=1 edge *)
				derivs,
				{
					N@D[function[u,v],u]/.{v->v0,u->0},
					N@D[function[u,v],u]/.{v->v1,u->1}
				},
				-2
			];
		];

		M=prvSplineInterpCoeffArrayD[m,p,U,K];
		polygon=
			Transpose[ (* polygon will need to be transposed since we are walking through column-wise *)
				Table[
					Q=prvSplineInterpDataVectorD[polygon[[All,i]],derivs[[i]],m,p,U];
					LinearSolve[M,Q],
					{i,1,vpts+3} (* fit to two extra stations *)
				]
			];

		NURBSSurface[Homogeneous[polygon],{U,V},{p,q}]
	]




End[]

EndPackage[]

